# coding: utf-8

"""
Intervals for luigi.DateParameter.

Usage:

    # run this on the 1st of every month ...
    date = luigi.DateParameter(date=intervals.monthly())

"""

import datetime

def daily(date=datetime.date.today()):
    """ just pass on the given date """
    return date

def weekly(date=datetime.date.today()):
    """ weeks start at Monday """
    return date - datetime.timedelta(days=date.weekday())

def biweekly(date=datetime.date.today()):
    """ every two weeks """
    return datetime.date(date.year, date.month, 1 if date.day < 15 else 15)

def monthly(date=datetime.date.today()):
    """ Take a date object and return the first day of the month. """
    return datetime.date(date.year, date.month, 1)

def quarterly(date=datetime.date.today()):
    """ 1/1, 4/1, 7/1, 10/1 """
    return datetime.date(date.year, ((date.month - 1)//3) * 3 + 1, 1)

def semiyearly(date=datetime.date.today()):
    return datetime.date(date.year, 1 if date.day < 7 else 7, 1)

def yearly(date=datetime.date.today()):
    return datetime.date(date.year, 1, 1)



# coding: utf-8

"""
Helper for databases.
"""

import logging

try:
    from MySQLdb.cursors import SSCursor
    import MySQLdb
except ImportError:
    logging.warn("MySQLdb seems missing: limited functionality.")

import sqlite3
import sqlitebck
import urlparse


class sqlite3db(object):
    """
    Simple context manager for sqlite3 databases. Commits everything at exit.

        with sqlite3db('/tmp/test.db') as cursor:
            query = cursor.execute('SELECT * FROM items')
            result = query.fetchall()

    For speedy, but memory hungry inserts you can first create a complete db
    in memory and then copy it to disk. 70000 INSERT/s have been observed, more
    is probably possible with faster IO.

        with sqlite3db(':memory:', copy_on_exit='/tmp/test.db') as cursor:
            cursor.execute("CREATE TABLE test (i INTEGER, t TEXT)")
            cursor.execute("INSERT INTO test VALUES (?, ?)",
                           (1, "Hello World"))

    """
    def __init__(self, path, copy_on_exit=None):
        self.path = path
        self.conn = None
        self.cursor = None
        self.copy_on_exit = copy_on_exit

    def __enter__(self):
        self.conn = sqlite3.connect(self.path)
        self.conn.text_factory = str
        self.cursor = self.conn.cursor()
        return self.cursor

    def __exit__(self, exc_class, exc, traceback):
        self.conn.commit()
        if self.copy_on_exit:
            target = sqlite3.connect(self.copy_on_exit)
            sqlitebck.copy(self.conn, target)
            target.close()
        self.conn.close()


class mysqldb(object):
    """ Context manager for MySQL database access.

        with mysqldb('mysql://user:pass@host/db', stream=True) as cursor:
            query = cursor.execute('SELECT * FROM items')
            result = query.fetchall()

    """
    def __init__(self, url, stream=False, commit_on_exit=False):
        result = urlparse.urlparse(url, scheme='mysql')
        self.hostname = result.hostname
        self.username = result.username
        self.password = result.password
        self.database = result.path.strip('/')
        self.stream = stream
        self.commit_on_exit = commit_on_exit
        self.conn = None
        self.cursor = None

    def __enter__(self):
        if self.stream:
            self.conn = MySQLdb.connect(host=self.hostname, user=self.username,
                                        passwd=self.password, db=self.database,
                                        cursorclass=SSCursor)
        else:
            self.conn = MySQLdb.connect(host=self.hostname, user=self.username,
                                        passwd=self.password, db=self.database)

        self.cursor = self.conn.cursor()
        return self.cursor

    def __exit__(self, exc_class, exc, traceback):
        if self.commit_on_exit:
            self.conn.commit()
        self.cursor.close()
        self.conn.close()



#!/usr/bin/env python
# coding: utf-8
# pylint: disable=C0301
"""
Parameter add-ons
=================

Custom luigi parameters.

"""
# pylint: disable=F0401,C0103,R0921,E1101
from luigi.parameter import (MissingParameterException, _no_default,
                             ParameterException)
import abc
import datetime
import luigi


_no_default = object()


class ILNParameter(luigi.Parameter):
    """
    Parse ILN (internal library number), so that the result is always in the
    *4 char* format.
    """
    def parse(self, s):
        """ Parse the value and pad left with zeroes. """
        return s.zfill(4)

class ClosestDateParameter(luigi.DateParameter):
    use_closest_date = True


class AbstractMappedDateParameter(luigi.DateParameter):
    """ A date parameter, that maps certain values to others on the fly.
    Useful, if the data source has non-standard update cycles, that can be
    evaluated on runtime. """

    # non-atomically increasing counter used for ordering parameters.
    counter = 0

    def __init__(self, default=_no_default, is_list=False, is_boolean=False,
                 is_global=False, significant=True, description=None,
                 default_from_config=None):
        """
        :param default: the default value for this parameter. This should match the type of the
                        Parameter, i.e. ``datetime.date`` for ``DateParameter`` or ``int`` for
                        ``IntParameter``. You may only specify either ``default`` or
                        ``default_from_config`` and not both. By default, no default is stored and
                        the value must be specified at runtime.
        :param bool is_list: specify ``True`` if the parameter should allow a list of values rather
                             than a single value. Default: ``False``. A list has an implicit default
                             value of ``[]``.
        :param bool is_boolean: specify ``True`` if the parameter is a boolean value. Default:
                                ``False``. Boolean's have an implicit default value of ``False``.
        :param bool is_global: specify ``True`` if the parameter is global (i.e. used by multiple
                               Tasks). Default: ``False``.
        :param bool significant: specify ``False`` if the parameter should not be treated as part of
                                 the unique identifier for a Task. An insignificant Parameter might
                                 also be used to specify a password or other sensitive information
                                 that should not be made public via the scheduler. Default:
                                 ``True``.
        :param str description: A human-readable string describing the purpose of this Parameter.
                                For command-line invocations, this will be used as the `help` string
                                shown to users. Default: ``None``.
        :param dict default_from_config: a dictionary with entries ``section`` and ``name``
                                         specifying a config file entry from which to read the
                                         default value for this parameter. You may only specify
                                         either ``default`` or ``default_from_config`` and not both.
                                         Default: ``None``.
        """
        # The default default is no default
        self.__default = default  # We also use this to store global values
        self.is_list = is_list
        self.is_boolean = is_boolean and not is_list  # Only BooleanParameter should ever use this. TODO(erikbern): should we raise some kind of exception?
        self.is_global = is_global  # It just means that the default value is exposed and you can override it
        self.significant = significant # Whether different values for this parameter will differentiate otherwise equal tasks
        if is_global and default == _no_default and default_from_config is None:
            raise ParameterException('Global parameters need default values')
        self.description = description

        if default != _no_default and default_from_config is not None:
            raise ParameterException('Can only specify either a default or a default_from_config')
        if default_from_config is not None and (not 'section' in default_from_config or not 'name' in default_from_config):
            raise ParameterException('default_from_config must be a hash containing entries for section and name')
        self.default_from_config = default_from_config

        # We need to keep track of this to get the order right (see Task class)
        self.counter = AbstractMappedDateParameter.counter
        AbstractMappedDateParameter.counter += 1


    @property
    def default(self):
        """The default value for this Parameter.

        :raises MissingParameterException: if a default is not set.
        :return: the parsed default value.
        """
        if self.__default == _no_default and self.default_from_config is None:
            raise MissingParameterException("No default specified")
        if self.__default != _no_default:
            try:
                return self.mapper(self.__default)
            except NotImplementedError:
                return self.__default

        value = self._get_default_from_config(safe=False)
        if self.is_list:
            return tuple(self.parse(p.strip()) for p in value.strip().split('\n'))
        else:
            return self.parse(value)

    def set_default(self, value):
        """Set the default value of this Parameter.

        :param value: the new default value.
        """
        self.__default = value

    @property
    def has_default(self):
        """``True`` if a default was specified or if default_from_config references a valid entry in the conf."""
        if self.default_from_config is not None:
            return self._get_default_from_config(safe=True) is not None
        return self.__default != _no_default

    @abc.abstractmethod
    def mapper(self, date):
        """ Must return the mapped date (object) for a given date (object).

            task = SomeOtherTask(date=date)
            luigi.build([task], local_scheduler=True)
            s = task.output().open().read().strip()
            return datetime.date(*(int(v) for v in s.split('-')))
        """
        raise NotImplementedError

    def parse(self, s):
        """ Parse value into date and map it. """
        date = datetime.date(*(int(v) for v in s.split('-')))
        return self.mapper(date)
